@@ -36,14 +36,6 @@ usage = (
     + ' [to [OUTFILE] [SAVE-OPTIONS]]'
 )
 
-
-def _get_for_location(registry, file, mode, format=''):
-    """Get loader/saver for font file location."""
-    if not file:
-        return registry.get_for(format=format)
-    with monobit.open_location(file, mode) as stream:
-        return registry.get_for(stream, format=format)
-
 def _get_context_help(rec):
     if rec.args:
         file = rec.args[0]
@@ -51,9 +43,9 @@ def _get_context_help(rec):
         file = rec.kwargs.get('infile', '')
     format = rec.kwargs.get('format', '')
     if rec.command == 'load':
-        func, *_ = _get_for_location(monobit.loaders, file, 'r', format)
+        func, *_ = monobit.loaders.get_for(format=format)
     else:
-        func, *_ = _get_for_location(monobit.savers, file, 'w', format)
+        func, *_ = monobit.savers.get_for(format=format)
     if func:
         return func.script_args
     return None
